@@ -18,9 +18,6 @@ import (
 	"bufio"
 	"context"
 	"fmt"
-	"io"
-	"log"
-	"net"
 	"reflect"
 	"strings"
 	"sync"
@@ -28,192 +25,9 @@ import (
 	"time"
 
 	irc "github.com/fluffle/goirc/client"
+	"github.com/google/alertmanager-irc-relay/logging"
 )
 
-type LineHandlerFunc func(*bufio.ReadWriter, *irc.Line) error
-
-func hUSER(conn *bufio.ReadWriter, line *irc.Line) error {
-	r := fmt.Sprintf(":example.com 001 %s :Welcome\n", line.Args[0])
-	_, err := conn.WriteString(r)
-	return err
-}
-
-func hQUIT(conn *bufio.ReadWriter, line *irc.Line) error {
-	return fmt.Errorf("client asked to terminate")
-}
-
-type closeEarlyHandler func()
-
-type testServer struct {
-	net.Listener
-	Client net.Conn
-
-	ServingWaitGroup     sync.WaitGroup
-	ConnectionsWaitGroup sync.WaitGroup
-
-	lineHandlersMu sync.Mutex
-	lineHandlers   map[string]LineHandlerFunc
-
-	Log []string
-
-	closeEarlyMu sync.Mutex
-	closeEarlyHandler
-}
-
-func (s *testServer) setDefaultHandlers() {
-	if s.lineHandlers == nil {
-		s.lineHandlers = make(map[string]LineHandlerFunc)
-	}
-	s.lineHandlers["USER"] = hUSER
-	s.lineHandlers["QUIT"] = hQUIT
-}
-
-func (s *testServer) getHandler(cmd string) LineHandlerFunc {
-	s.lineHandlersMu.Lock()
-	defer s.lineHandlersMu.Unlock()
-	return s.lineHandlers[cmd]
-}
-
-func (s *testServer) SetHandler(cmd string, h LineHandlerFunc) {
-	s.lineHandlersMu.Lock()
-	defer s.lineHandlersMu.Unlock()
-	if h == nil {
-		delete(s.lineHandlers, cmd)
-	} else {
-		s.lineHandlers[cmd] = h
-	}
-}
-
-func (s *testServer) handleLine(conn *bufio.ReadWriter, line *irc.Line) error {
-	s.Log = append(s.Log, strings.Trim(line.Raw, " \r\n"))
-	handler := s.getHandler(line.Cmd)
-	if handler == nil {
-		log.Printf("=Server= No handler for command '%s', skipping", line.Cmd)
-		return nil
-	}
-	return handler(conn, line)
-}
-
-func (s *testServer) handleConnection(conn net.Conn) {
-	defer func() {
-		s.Client = nil
-		conn.Close()
-		s.ConnectionsWaitGroup.Done()
-	}()
-	bufConn := bufio.NewReadWriter(bufio.NewReader(conn), bufio.NewWriter(conn))
-	for {
-		msg, err := bufConn.ReadBytes('\n')
-		if err != nil {
-			if err == io.EOF {
-				log.Printf("=Server= Client %s disconnected", conn.RemoteAddr().String())
-			} else {
-				log.Printf("=Server= Could not read from %s: %s", conn.RemoteAddr().String(), err)
-			}
-			return
-		}
-		log.Printf("=Server= Received %s", msg)
-		line := irc.ParseLine(string(msg))
-		if line == nil {
-			log.Printf("=Server= Could not parse received line")
-			continue
-		}
-		err = s.handleLine(bufConn, line)
-		if err != nil {
-			log.Printf("=Server= Closing connection: %s", err)
-			return
-		}
-		bufConn.Flush()
-	}
-}
-
-func (s *testServer) SetCloseEarly(h closeEarlyHandler) {
-	s.closeEarlyMu.Lock()
-	defer s.closeEarlyMu.Unlock()
-	s.closeEarlyHandler = h
-}
-
-func (s *testServer) handleCloseEarly(conn net.Conn) bool {
-	s.closeEarlyMu.Lock()
-	defer s.closeEarlyMu.Unlock()
-	if s.closeEarlyHandler == nil {
-		return false
-	}
-	log.Printf("=Server= Closing connection early")
-	conn.Close()
-	s.closeEarlyHandler()
-	return true
-}
-
-func (s *testServer) Serve() {
-	defer s.ServingWaitGroup.Done()
-	for {
-		conn, err := s.Listener.Accept()
-		if err != nil {
-			log.Printf("=Server= Stopped accepting new connections")
-			return
-		}
-		log.Printf("=Server= New client connected from %s", conn.RemoteAddr().String())
-		if s.handleCloseEarly(conn) {
-			continue
-		}
-		s.Client = conn
-		s.ConnectionsWaitGroup.Add(1)
-		s.handleConnection(conn)
-	}
-}
-
-func (s *testServer) Stop() {
-	s.Listener.Close()
-	s.ServingWaitGroup.Wait()
-	s.ConnectionsWaitGroup.Wait()
-}
-
-func makeTestServer(t *testing.T) (*testServer, int) {
-	server := new(testServer)
-	server.Log = make([]string, 0)
-	server.setDefaultHandlers()
-
-	addr, err := net.ResolveTCPAddr("tcp", "127.0.0.1:0")
-	if err != nil {
-		t.Fatalf("=Server= Could not resolve tcp addr: %s", err)
-	}
-	listener, err := net.ListenTCP("tcp", addr)
-	if err != nil {
-		t.Fatalf("=Server= Could not create listener: %s", err)
-	}
-	addr = listener.Addr().(*net.TCPAddr)
-	log.Printf("=Server= Test server listening on %s", addr.String())
-
-	server.Listener = listener
-
-	server.ServingWaitGroup.Add(1)
-	go func() {
-		server.Serve()
-	}()
-
-	addr = listener.Addr().(*net.TCPAddr)
-	return server, addr.Port
-}
-
-type FakeDelayer struct {
-	DelayOnChan bool
-	StopDelay   chan bool
-}
-
-func (f *FakeDelayer) Delay() {
-	f.DelayContext(context.Background())
-}
-
-func (f *FakeDelayer) DelayContext(ctx context.Context) bool {
-	log.Printf("Faking Backoff")
-	if f.DelayOnChan {
-		log.Printf("Waiting StopDelay signal")
-		<-f.StopDelay
-		log.Printf("Received StopDelay signal")
-	}
-	return true
-}
-
 func makeTestIRCConfig(IRCPort int) *Config {
 	return &Config{
 		IRCNick:     "foo",
@@ -223,113 +37,74 @@ func makeTestIRCConfig(IRCPort int) *Config {
 		IRCUseSSL:   false,
 		IRCChannels: []IRCChannel{
 			IRCChannel{Name: "#foo"},
-			IRCChannel{Name: "#bar"},
-			IRCChannel{Name: "#baz"},
 		},
 		UsePrivmsg: false,
+		NickservIdentifyPatterns: []string{
+			"identify yourself ktnxbye",
+		},
 	}
 }
 
-func makeTestNotifier(t *testing.T, config *Config) (*IRCNotifier, chan AlertMsg, context.CancelFunc, *sync.WaitGroup) {
+func makeTestNotifier(t *testing.T, config *Config) (*IRCNotifier, chan AlertMsg, context.Context, context.CancelFunc, *sync.WaitGroup) {
+	fakeDelayerMaker := &FakeDelayerMaker{}
+	fakeTime := &FakeTime{
+		afterChan: make(chan time.Time, 1),
+	}
 	alertMsgs := make(chan AlertMsg)
 	ctx, cancel := context.WithCancel(context.Background())
 	stopWg := sync.WaitGroup{}
 	stopWg.Add(1)
-	notifier, err := NewIRCNotifier(ctx, &stopWg, config, alertMsgs)
+	notifier, err := NewIRCNotifier(config, alertMsgs, fakeDelayerMaker, fakeTime)
 	if err != nil {
 		t.Fatal(fmt.Sprintf("Could not create IRC notifier: %s", err))
 	}
 	notifier.Client.Config().Flood = true
-	notifier.BackoffCounter = &FakeDelayer{
-		DelayOnChan: false,
-		StopDelay:   make(chan bool),
-	}
 
-	return notifier, alertMsgs, cancel, &stopWg
-}
-
-func TestPreJoinChannels(t *testing.T) {
-	server, port := makeTestServer(t)
-	config := makeTestIRCConfig(port)
-	notifier, _, cancel, _ := makeTestNotifier(t, config)
-
-	var testStep sync.WaitGroup
-
-	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			testStep.Done()
-		}
-		return nil
-	}
-	server.SetHandler("JOIN", joinHandler)
-
-	testStep.Add(1)
-	go notifier.Run()
-
-	testStep.Wait()
-
-	cancel()
-	server.Stop()
-
-	expectedCommands := []string{
-		"NICK foo",
-		"USER foo 12 * :",
-		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
-		"QUIT :see ya",
-	}
-
-	if !reflect.DeepEqual(expectedCommands, server.Log) {
-		t.Error("Did not pre-join channels")
-	}
+	return notifier, alertMsgs, ctx, cancel, &stopWg
 }
 
 func TestServerPassword(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
 	config.IRCHostPass = "hostsecret"
-	notifier, _, cancel, _ := makeTestNotifier(t, config)
+	notifier, _, ctx, cancel, stopWg := makeTestNotifier(t, config)
 
 	var testStep sync.WaitGroup
 
 	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			testStep.Done()
-		}
+		testStep.Done()
 		return nil
 	}
 	server.SetHandler("JOIN", joinHandler)
 
 	testStep.Add(1)
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		"PASS hostsecret",
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		"QUIT :see ya",
 	}
 
 	if !reflect.DeepEqual(expectedCommands, server.Log) {
-		t.Error("Did not send IRC server password")
+		t.Error("Did not send IRC server password. Received commands:\n", strings.Join(server.Log, "\n"))
 	}
 }
 
 func TestSendAlertOnPreJoinedChannel(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
-	notifier, alertMsgs, cancel, _ := makeTestNotifier(t, config)
+	notifier, alertMsgs, ctx, cancel, stopWg := makeTestNotifier(t, config)
 
 	var testStep sync.WaitGroup
 
@@ -342,16 +117,16 @@ func TestSendAlertOnPreJoinedChannel(t *testing.T) {
 		if line.Args[0] == testChannel {
 			testStep.Done()
 		}
-		return nil
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinedHandler)
 
 	testStep.Add(1)
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	testStep.Wait()
 
-	server.SetHandler("JOIN", nil)
+	server.SetHandler("JOIN", hJOIN)
 
 	noticeHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
 		testStep.Done()
@@ -365,14 +140,15 @@ func TestSendAlertOnPreJoinedChannel(t *testing.T) {
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		"NOTICE #foo :test message",
 		"QUIT :see ya",
 	}
@@ -386,7 +162,7 @@ func TestUsePrivmsgToSendAlertOnPreJoinedChannel(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
 	config.UsePrivmsg = true
-	notifier, alertMsgs, cancel, _ := makeTestNotifier(t, config)
+	notifier, alertMsgs, ctx, cancel, stopWg := makeTestNotifier(t, config)
 
 	var testStep sync.WaitGroup
 
@@ -399,16 +175,16 @@ func TestUsePrivmsgToSendAlertOnPreJoinedChannel(t *testing.T) {
 		if line.Args[0] == testChannel {
 			testStep.Done()
 		}
-		return nil
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinedHandler)
 
 	testStep.Add(1)
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	testStep.Wait()
 
-	server.SetHandler("JOIN", nil)
+	server.SetHandler("JOIN", hJOIN)
 
 	privmsgHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
 		testStep.Done()
@@ -422,14 +198,15 @@ func TestUsePrivmsgToSendAlertOnPreJoinedChannel(t *testing.T) {
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		"PRIVMSG #foo :test message",
 		"QUIT :see ya",
 	}
@@ -442,7 +219,7 @@ func TestUsePrivmsgToSendAlertOnPreJoinedChannel(t *testing.T) {
 func TestSendAlertAndJoinChannel(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
-	notifier, alertMsgs, cancel, _ := makeTestNotifier(t, config)
+	notifier, alertMsgs, ctx, cancel, stopWg := makeTestNotifier(t, config)
 
 	var testStep sync.WaitGroup
 
@@ -452,20 +229,17 @@ func TestSendAlertAndJoinChannel(t *testing.T) {
 	// Send the alert after configured channels have joined, to ensure log
 	// ordering.
 	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			testStep.Done()
-		}
-		return nil
+		testStep.Done()
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinHandler)
 
 	testStep.Add(1)
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	testStep.Wait()
 
-	server.SetHandler("JOIN", nil)
+	server.SetHandler("JOIN", hJOIN)
 
 	noticeHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
 		testStep.Done()
@@ -479,15 +253,17 @@ func TestSendAlertAndJoinChannel(t *testing.T) {
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		// #foobar joined before sending message
+		"PRIVMSG ChanServ :UNBAN #foobar",
 		"JOIN #foobar",
 		"NOTICE #foobar :test message",
 		"QUIT :see ya",
@@ -501,7 +277,7 @@ func TestSendAlertAndJoinChannel(t *testing.T) {
 func TestSendAlertDisconnected(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
-	notifier, alertMsgs, cancel, _ := makeTestNotifier(t, config)
+	notifier, alertMsgs, ctx, cancel, stopWg := makeTestNotifier(t, config)
 
 	var testStep, holdUserStep sync.WaitGroup
 
@@ -513,15 +289,15 @@ func TestSendAlertDisconnected(t *testing.T) {
 	testStep.Add(1)
 	holdUserStep.Add(1)
 	holdUser := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		log.Printf("=Server= Wait before completing session")
+		logging.Info("=Server= Wait before completing session")
 		testStep.Wait()
-		log.Printf("=Server= Completing session")
+		logging.Info("=Server= Completing session")
 		holdUserStep.Done()
 		return hUSER(conn, line)
 	}
 	server.SetHandler("USER", holdUser)
 
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	// Alert channels is not consumed while disconnected
 	select {
@@ -534,14 +310,11 @@ func TestSendAlertDisconnected(t *testing.T) {
 	holdUserStep.Wait()
 
 	// Make sure session is established by checking that pre-joined
-	// channels are there.
+	// channel is there.
 	testStep.Add(1)
 	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			testStep.Done()
-		}
-		return nil
+		testStep.Done()
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinHandler)
 
@@ -560,14 +333,15 @@ func TestSendAlertDisconnected(t *testing.T) {
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		// Only message sent while being connected is received.
 		"NOTICE #foo :connected test message",
 		"QUIT :see ya",
@@ -581,48 +355,45 @@ func TestSendAlertDisconnected(t *testing.T) {
 func TestReconnect(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
-	notifier, _, cancel, _ := makeTestNotifier(t, config)
+	notifier, _, ctx, cancel, stopWg := makeTestNotifier(t, config)
 
 	var testStep sync.WaitGroup
 
 	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			testStep.Done()
-		}
-		return nil
+		testStep.Done()
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinHandler)
 
 	testStep.Add(1)
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
-	// Wait until the last pre-joined channel is seen.
+	// Wait until the pre-joined channel is seen.
 	testStep.Wait()
 
 	// Simulate disconnection.
 	testStep.Add(1)
 	server.Client.Close()
 
-	// Wait again until the last pre-joined channel is seen.
+	// Wait again until the pre-joined channel is seen.
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		// Commands from first connection
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		// Commands from reconnection
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		"QUIT :see ya",
 	}
 
@@ -637,7 +408,7 @@ func TestConnectErrorRetry(t *testing.T) {
 	// Attempt SSL handshake. The server does not support it, resulting in
 	// a connection error.
 	config.IRCUseSSL = true
-	notifier, _, cancel, _ := makeTestNotifier(t, config)
+	notifier, _, ctx, cancel, stopWg := makeTestNotifier(t, config)
 	// Pilot reconnect attempts via backoff delay to prevent race
 	// conditions in the test while we change the components behavior on
 	// the fly.
@@ -653,7 +424,7 @@ func TestConnectErrorRetry(t *testing.T) {
 
 	server.SetCloseEarly(earlyHandler)
 
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	delayer.StopDelay <- true
 
@@ -663,11 +434,8 @@ func TestConnectErrorRetry(t *testing.T) {
 	notifier.Client.Config().SSL = false
 	joinStep.Add(1)
 	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			joinStep.Done()
-		}
-		return nil
+		joinStep.Done()
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinHandler)
 	server.SetCloseEarly(nil)
@@ -677,14 +445,15 @@ func TestConnectErrorRetry(t *testing.T) {
 	joinStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		"NICK foo",
 		"USER foo 12 * :",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		"QUIT :see ya",
 	}
 
@@ -697,37 +466,44 @@ func TestIdentify(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
 	config.IRCNickPass = "nickpassword"
-	notifier, _, cancel, _ := makeTestNotifier(t, config)
+	notifier, _, ctx, cancel, stopWg := makeTestNotifier(t, config)
 	notifier.NickservDelayWait = 0 * time.Second
 
 	var testStep sync.WaitGroup
 
-	// Wait until the last pre-joined channel is seen (joining happens
+	// Trigger NickServ identify request when we see the NICK command
+	// Note: We also test formatting cleanup with this message
+	nickHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
+		var err error
+		_, err = conn.WriteString(":NickServ!NickServ@services. NOTICE airtest :This nickname is registered. Please choose a different nickname, or \002identify yourself\002 ktnxbye.\n")
+		return err
+	}
+	server.SetHandler("NICK", nickHandler)
+
+	// Wait until the pre-joined channel is seen (joining happens
 	// after identification).
 	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			testStep.Done()
-		}
-		return nil
+		testStep.Done()
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinHandler)
 
 	testStep.Add(1)
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
 		"NICK foo",
 		"USER foo 12 * :",
 		"PRIVMSG NickServ :IDENTIFY nickpassword",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		"QUIT :see ya",
 	}
 
@@ -736,50 +512,51 @@ func TestIdentify(t *testing.T) {
 	}
 }
 
-func TestGhostAndIdentify(t *testing.T) {
+func TestGhost(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
 	config.IRCNickPass = "nickpassword"
-	notifier, _, cancel, _ := makeTestNotifier(t, config)
+	notifier, _, ctx, cancel, stopWg := makeTestNotifier(t, config)
 	notifier.NickservDelayWait = 0 * time.Second
 
-	var testStep, usedNick, unregisteredNickHandler sync.WaitGroup
+	var testStep sync.WaitGroup
 
-	// Trigger 433 for first nick
-	usedNick.Add(1)
-	unregisteredNickHandler.Add(1)
-	nickHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
+	// Trigger 433 for first nick when we see the USER command
+	userHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
 		var err error
 		if line.Args[0] == "foo" {
 			_, err = conn.WriteString(":example.com 433 * foo :nick in use\n")
 		}
-		usedNick.Done()
-		unregisteredNickHandler.Wait()
+		return err
+	}
+	server.SetHandler("USER", userHandler)
+
+	// Trigger 001 when we see NICK foo^
+	nickHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
+		var err error
+		if line.Args[0] == "foo^" {
+			_, err = conn.WriteString(":example.com 001 foo^ :Welcome\n")
+		}
 		return err
 	}
 	server.SetHandler("NICK", nickHandler)
 
-	// Wait until the last pre-joined channel is seen (joining happens
-	// after identification).
+	// Wait until the pre-joined channel is seen (joining happens
+	// after ghosting).
 	joinHandler := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		// #baz is configured as the last channel to pre-join
-		if line.Args[0] == "#baz" {
-			testStep.Done()
-		}
-		return nil
+		testStep.Done()
+		return hJOIN(conn, line)
 	}
 	server.SetHandler("JOIN", joinHandler)
 
 	testStep.Add(1)
-	go notifier.Run()
-
-	usedNick.Wait()
-	server.SetHandler("NICK", nil)
-	unregisteredNickHandler.Done()
+	go notifier.Run(ctx, stopWg)
 
 	testStep.Wait()
 
 	cancel()
+	stopWg.Wait()
+
 	server.Stop()
 
 	expectedCommands := []string{
@@ -788,10 +565,8 @@ func TestGhostAndIdentify(t *testing.T) {
 		"NICK foo^",
 		"PRIVMSG NickServ :GHOST foo nickpassword",
 		"NICK foo",
-		"PRIVMSG NickServ :IDENTIFY nickpassword",
+		"PRIVMSG ChanServ :UNBAN #foo",
 		"JOIN #foo",
-		"JOIN #bar",
-		"JOIN #baz",
 		"QUIT :see ya",
 	}
 
@@ -803,48 +578,32 @@ func TestGhostAndIdentify(t *testing.T) {
 func TestStopRunningWhenHalfConnected(t *testing.T) {
 	server, port := makeTestServer(t)
 	config := makeTestIRCConfig(port)
-	notifier, _, cancel, stopWg := makeTestNotifier(t, config)
+	notifier, _, ctx, cancel, stopWg := makeTestNotifier(t, config)
 
-	var testStep, holdQuitWait sync.WaitGroup
+	var testStep sync.WaitGroup
 
 	// Send a StopRunning request while the client is connected but the
 	// session is not up
 	testStep.Add(1)
 	holdUser := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		log.Printf("=Server= NOT completing session")
+		logging.Info("=Server= NOT completing session")
 		testStep.Done()
 		return nil
 	}
 	server.SetHandler("USER", holdUser)
 
-	// Ignore quit, but wait for it to have deterministic test commands
-	holdQuitWait.Add(1)
-	holdQuit := func(conn *bufio.ReadWriter, line *irc.Line) error {
-		log.Printf("=Server= Ignoring quit")
-		holdQuitWait.Done()
-		return nil
-	}
-	server.SetHandler("QUIT", holdQuit)
-
-	go notifier.Run()
+	go notifier.Run(ctx, stopWg)
 
 	testStep.Wait()
 
 	cancel()
-
 	stopWg.Wait()
 
-	holdQuitWait.Wait()
-
-	// Client has left, cleanup the server side before stopping
-	server.Client.Close()
-
 	server.Stop()
 
 	expectedCommands := []string{
 		"NICK foo",
 		"USER foo 12 * :",
-		"QUIT :see ya",
 	}
 
 	if !reflect.DeepEqual(expectedCommands, server.Log) {
